/* ===================================================================================
 * Class to interface with ADXL345 sensor boards/chips
 * ===================================================================================
 * Author: Thomas D. Kryger
 * ===================================================================================
  * You may use, copy, borrow, modify or do anything you like with this code. Crediting
 * the author will be appreciated.
 * ===================================================================================
 */
using System;
using LoveElectronics.Resources;
using Microsoft.SPOT.Hardware;
using NETMF.IO.I2C;

namespace SensorTest.SensorDrivers
{
    /// <summary>
    /// Encapsulates communication and management of the Analog Devices ADXL345 3-axis accelerometer.
    /// Datasheet: http://www.analog.com/static/imported-files/data_sheets/ADXL345.pdf
    /// Tutorial: https://www.loveelectronics.co.uk/Tutorials/13/tilt-compensated-compass-arduino-tutorial
    /// Tutorial: https://www.loveelectronics.co.uk/Tutorials/10/accelerometers-and-how-to-use-them
    /// </summary>
    public class ADXL345 : AbstractI2CDevice
    {
        #region constructor
        public ADXL345(ushort address, int clockRate, int timeout, Cpu.Pin InterruptPort)
            : base(address, clockRate, timeout)
        {
            if (!Connected())
                throw new Exception("ADXL345 not connected");
            if (InterruptPort != Cpu.Pin.GPIO_NONE)
            {
                interruptPort = new InterruptPort(InterruptPort, true, Port.ResistorMode.Disabled, Port.InterruptMode.InterruptEdgeHigh);
            }
            Range = 2;
            FullResolution = true;
            SetDataRate(ADXL345.DataRates.DataRate50Hz);
            EnableMeasurements();
            if (interruptPort != null)
                interruptPort.OnInterrupt += new NativeEventHandler(interruptPort_OnInterrupt);
        }

        #endregion

        #region Delegates and events
        public delegate void ADXL345NewData(uint data1, uint data2, DateTime time);
        public event ADXL345NewData OnNewData;

        void interruptPort_OnInterrupt(uint data1, uint data2, DateTime time)
        {
            if (OnNewData != null)
                OnNewData(data1, data2, time);
        }
        #endregion

        #region Consts
        #region Scale Factors
        private const double OffsetMgScale = 62.5d;
        /// <summary>
        /// The scaling to use if the device is not in Full Resolution mode and has a Range of 2g.
        /// </summary>
        public const double ScaleFor2G_mg = 3.9d;
        public const double ScaleFor2G_g = 0.0039d;
        /// <summary>
        /// The scaling to use if the device is not in Full Resolution mode and has a Range of 4g.
        /// </summary>
        public const double ScaleFor4G_mg = 7.8d;
        public const double ScaleFor4G_g = 0.0078d;
        /// <summary>
        /// The scaling to use if the device is not in Full Resolution mode and has a Range of 8g.
        /// </summary>
        public const double ScaleFor8G_mg = 15.6d;
        public const double ScaleFor8G_g = 0.0156d;
        /// <summary>
        /// The scaling to use if the device is not in Full Resolution mode and has a Range of 16g.
        /// </summary>
        public const double ScaleFor16G_mg = 31.2d;
        public const double ScaleFor16G_g = 0.0312d;
        #endregion
        #endregion

        #region enums
        /// <summary>
        /// Defines the types of events the device raises.
        /// </summary>
        [Flags]
        public enum EventType : byte
        {
            DataReady = 0x80,
            SingleTap = 0x40,
            DoubleTap = 0x20,
            Activity = 0x10,
            Inactivity = 0x08,
            FreeFall = 0x04,
            Watermark = 0x02,
            Overrun = 0x01
        }
        public enum DataRates : byte
        {
            DataRate1600Hz = 0x0F,
            DataRate800Hz = 0x0E,
            DataRate400Hz = 0x0D,
            DataRate200Hz = 0x0C,
            DataRate100Hz = 0x0B,
            DataRate50Hz = 0x0A,
            DataRate25Hz = 0x09,
            DataRate12_5Hz = 0x08,
            DataRate6_25Hz = 0x07,
            DataRate3_13Hz = 0x06,
            DataRate1_56Hz = 0x05,
            DataRate0_78Hz = 0x04,
            DataRate0_39Hz = 0x03,
            DataRate0_25Hz = 0x02,
            DataRate0_10Hz = 0x01,
            DataRate0_05Hz = 0x00
        }
        public enum Register : byte
        {
            DeviceId = 0x00,
            TapThreshold = 0x1D,
            OffsetX = 0x1E,
            OffsetY = 0x1F,
            OffsetZ = 0x20,
            TapDuration = 0x21,
            TapLatency = 0x22,
            TapWindow = 0x23,
            ActivityThreshold = 0x24,
            InactivityThreshold = 0x25,
            InactivityTime = 0x26,
            /// <summary>
            /// Axis enable control for activity and inactivity detection
            /// -- From datasheet.
            /// </summary>
            ACT_INACT_CTL = 0x27,
            FreeFallThreshold = 0x28,
            FreeFallTime = 0x29,
            /// <summary>
            /// Axis control for single tap/double tab.
            /// </summary>
            TapAxis = 0x2A,
            /// <summary>
            /// Source of single tab/double tab.
            /// </summary>
            TapSources = 0x2B,
            /// <summary>
            /// Data Rate and power mode control.
            /// </summary>
            BandwidthRate = 0x2C,
            /// <summary>
            /// Power saving features control.
            /// </summary>
            PowerControl = 0x2D,
            InterruptEnable = 0x2E,
            InterruptMapping = 0x2F,
            InterruptSource = 0x30,
            DataFormat = 0x31,

            DataX_0 = 0x32,
            DataX_1 = 0x33,
            DataY_0 = 0x34,
            DataY_1 = 0x35,
            DataZ_0 = 0x36,
            DataZ_1 = 0x37,

            FifoControl = 0x56,
            FifoStatus = 0x57,
            // Bit 3 of the PowerControl register specifies measurement mode.
            PowerControl_MeasurementsEnabledBit = 0x08,
            // Bit 4 of the PowerControl register specifices auto sleep.
            PowerControl_MeasurementsAutoSleepBit = 0x04
        }
        #endregion

        #region fields
        private InterruptPort interruptPort;
        #endregion

        #region Properties
        /// <summary>
        /// Returns the most recently retreived X axis reading in mG.
        /// Only use this if the device is in Full Resolution mode, otherwise scale the result yourself with this.ScaleForXG
        /// </summary>
        public double ScaledXAxis { get { return X * ScaleFor2G_mg; } }

        /// <summary>
        /// Returns the most recently retreived Y axis reading in mG.
        /// Only use this if the device is in Full Resolution mode, otherwise scale the result yourself with this.ScaleForXG
        /// </summary>
        public double ScaledYAxis { get { return Y * ScaleFor2G_mg; } }

        /// <summary>
        /// Returns the most recently retreived Z axis reading in mG.
        /// Only use this if the device is in Full Resolution mode, otherwise scale the result yourself with this.ScaleForXG
        /// </summary>
        public double ScaledZAxis { get { return Z * ScaleFor2G_mg; } }


        /// <summary>
        /// Returns the most recently retreived X axis reading in G.
        /// Only use this if the device is in Full Resolution mode, otherwise scale the result yourself with this.ScaleForXG
        /// </summary>
        public double ScaledXAxisG { get { return X * ScaleFor2G_g; } }

        /// <summary>
        /// Returns the most recently retreived Y axis reading in G.
        /// Only use this if the device is in Full Resolution mode, otherwise scale the result yourself with this.ScaleForXG
        /// </summary>
        public double ScaledYAxisG { get { return Y * ScaleFor2G_g; } }

        /// <summary>
        /// Returns the most recently retreived Z axis reading in G.
        /// Only use this if the device is in Full Resolution mode, otherwise scale the result yourself with this.ScaleForXG
        /// </summary>
        public double ScaledZAxisG { get { return Z * ScaleFor2G_g; } }

        /// <summary>
        /// Returns the X axis reading.
        /// </summary>
        public double X { get { return ReadShort((byte)Register.DataX_1, (byte)Register.DataX_0); } }

        /// <summary>
        /// Returns the Y axis reading.
        /// </summary>
        public double Y { get { return ReadShort((byte)Register.DataY_1, (byte)Register.DataY_0); } }

        /// <summary>
        /// Returns the Z axis reading.
        /// </summary>
        public double Z { get { return ReadShort((byte)Register.DataZ_1, (byte)Register.DataZ_0); } }


        /// <summary>
        /// Get/set the offset of the X axis in mG.
        /// </summary>
        public double ScaledOffsetX
        {
            get { return OffsetX * OffsetMgScale; }
            set { OffsetX = (byte)(value / OffsetMgScale); }
        }

        /// <summary>
        /// Get/set the offset of the Y axis in mG.
        /// </summary>
        public double ScaledOffsetY
        {
            get { return OffsetY * OffsetMgScale; }
            set { OffsetY = (byte)(value / OffsetMgScale); }
        }

        /// <summary>
        /// Get/set the offset of the Z axis in mG.
        /// </summary>
        public double ScaledOffsetZ
        {
            get { return OffsetZ * OffsetMgScale; }
            set { OffsetZ = (byte)(value / OffsetMgScale); }
        }


        /// <summary>
        /// Get/set the raw offset of the X axis.
        /// </summary>
        public byte OffsetX
        {
            get { return Read((byte)Register.OffsetX)[0]; }
            set { Write((byte)Register.OffsetX, value); }
        }

        /// <summary>
        /// Get/set the raw offset of the Y axis.
        /// </summary>
        public byte OffsetY
        {
            get { return Read((byte)Register.OffsetY)[0]; }
            set { Write((byte)Register.OffsetY, value); }
        }

        /// <summary>
        /// Get/set the raw offset of the Z axis.
        /// </summary>
        public byte OffsetZ
        {
            get { return Read((byte)Register.OffsetZ)[0]; }
            set { Write((byte)Register.OffsetZ, value); }
        }

        /// <summary>
        /// Range of device in G
        /// </summary>
        /// <remarks>Must be either 2, 4, 8 or 16g.</remarks>
        public int Range
        {
            get
            {
                byte data = Read((byte)Register.DataFormat)[0];
                if ((data & 0x03) == 0x03)
                    return 16;
                if ((data & 0x02) == 0x02)
                    return 8;
                if ((data & 0x01) == 0x01)
                    return 4;
                else
                    return 2;
            }
            set
            {
                byte data = Read((byte)Register.DataFormat)[0];
                switch (value)
                {
                    case 2:
                        unchecked { data &= (byte)~0x03; }
                        break;
                    case 4:
                        data |= 0x01;
                        unchecked { data &= (byte)~0x02; }
                        break;
                    case 8:
                        data |= 0x02;
                        unchecked { data &= (byte)~0x01; }
                        break;
                    case 16:
                        data |= 0x03;
                        break;
                }
                Write((byte)Register.DataFormat, data);
            }
        }

        /// <summary>
        /// In full resolution mode the output resolution increases with the g range to maintain 4mg/LSB scale.
        /// When in not full resolution the device is in 10bit mode, and the Range determins max g range and scale.
        /// </summary>
        public bool FullResolution
        {
            get
            {
                byte data = Read((byte)Register.DataFormat)[0];
                if ((data & 0x08) == 0x08)
                    return true;
                else
                    return false;
            }
            set
            {
                byte data = Read((byte)Register.DataFormat)[0];
                if (value)
                    data |= 0x08;
                else
                    unchecked { data &= (byte)~0x08; }
                Write((byte)Register.DataFormat, data);
            }
        }

        /// <summary>
        /// Pitch (Y) in radians
        /// </summary>
        public double Pitch { get { return ExMath.Asin(FixOneToMinus1(ScaledYAxisG)); } }

        /// <summary>
        /// Roll (X) in radians
        /// </summary>
        public double Roll { get { return ExMath.Asin(FixOneToMinus1(ScaledXAxisG)); } }

        /// <summary>
        /// Yaw (Z) in radians
        /// </summary>
        public double Yaw { get { return ExMath.Asin(FixOneToMinus1(ScaledZAxisG)); } }

        #endregion

        #region Sensor Control
        /// <summary>
        /// Sets the rate at which data is collected, see datasheet for values.
        /// </summary>
        /// <param name="rateCode"></param>
        public void SetDataRate(DataRates dr)
        {
            byte rateCode = (byte)dr;
            rateCode |= 0x10;
            Write((byte)Register.BandwidthRate, rateCode);
        }

        /// <summary>
        /// Enables measurements on the device.
        /// </summary>
        public void EnableMeasurements()
        {
            Write((byte)Register.PowerControl, (byte)Register.PowerControl_MeasurementsEnabledBit);
        }


        /// <summary>
        ///  Queries the device to see if it is in measurement mode.
        /// </summary>
        /// <returns></returns>
        public bool AreMeasurementsEnabled()
        {
            byte data = Read((byte)Register.PowerControl)[0];
            return ((data & (byte)Register.PowerControl_MeasurementsEnabledBit) == (byte)Register.PowerControl_MeasurementsEnabledBit);
        }


        /// <summary>
        /// Sets AutoSleep
        /// </summary>
        /// <param name="on"></param>
        public void SetAutoSleep(bool autoSleepOn)
        {
            byte data = Read((byte)Register.PowerControl)[0];
            if (autoSleepOn)
                data |= (byte)Register.PowerControl_MeasurementsAutoSleepBit;
            else
                unchecked
                {
                    data &= (byte)(~(byte)Register.PowerControl_MeasurementsAutoSleepBit);
                }
            // If we are clearing AutoSleep and we are in Measurement Mode, we must put the Device in and out of Standby.
            if (!autoSleepOn && (data & (byte)Register.PowerControl_MeasurementsEnabledBit) == (byte)Register.PowerControl_MeasurementsEnabledBit)
            {
                Standby();
                EnableMeasurements();
            }
        }


        /// <summary>
        /// Places the device in standby where it uses minimal current, there will be no data available during this time.
        /// </summary>
        public void Standby()
        {
            byte data = Read((byte)Register.PowerControl)[0];
            unchecked { data &= (byte)(~(byte)Register.PowerControl_MeasurementsEnabledBit); }
            Write((byte)Register.PowerControl, data);
        }

        /// <summary>
        /// Sets the time the sensor must be in free fall before raising a free fall event.
        /// </summary>
        /// <param name="fallMs"></param>
        public void SetFreeFallTime(int fallMs)
        {
            byte data = (byte)(fallMs / 5);
            Write((byte)Register.FreeFallTime, data);
        }


        /// <summary>
        /// Sets the mG that the sensor must not exceed to qualify for free fall.
        /// </summary>
        /// <param name="mg"></param>
        public void SetFreeFallThreshold(int mg)
        {
            byte data = (byte)(mg / 62.5);
            Write((byte)Register.FreeFallThreshold, data);
        }
        #endregion

        #region private methods
        /// <summary>
        /// Converts a byte array into a short. Taken from the BitConverter in the community.
        /// </summary>
        /// <param name="value"></param>
        /// <param name="index">JCarter</param>
        /// <returns></returns>
        private short ToInt16(byte[] value, int index = 0)
        {
            return (short)(
                value[0 + index] << 0 |
                value[1 + index] << 8);
        }

        private double FixOneToMinus1(double d)
        {
            if (d > 1)
                d = 1;
            else
                if (d < -1)
                    d = -1;
            return d;
        }
        #endregion

        public override bool Connected()
        {
            byte me = DeviceIdentifier()[0];
            return (me == 0xE5);
        }

        public override byte[] DeviceIdentifier()
        {
            return new byte[1] { Read((byte)Register.DeviceId)[0] };
        }
    }
}
