/* ===================================================================================
 * Class to interface with BMP085 sensor boards/chips
 * ===================================================================================
 * Author: Thomas D. Kryger
 * ===================================================================================
  * You may use, copy, borrow, modify or do anything you like with this code. Crediting
 * the author will be appreciated.
 * ===================================================================================
 */
using System;
using System.Threading;
using Microsoft.SPOT.Hardware;
using NETMF.IO.I2C;

namespace SensorTest.SensorDrivers
{
    /// <summary>
    /// The BMP085 is a high-precision, ultra-low power barometric pressure sensor. It offers a measuring range of 300 to 1100 hPa with an absolute accuracy of down to 0.03 hPa. The BMP085 is based on piezo-resistive technology for EMC robustness, high accuracy and linearity as well as long term stability.
    /// Datasheet: http://www.sparkfun.com/datasheets/Components/General/BST-BMP085-DS000-05.pdf
    /// </summary>
    public class BMP085 : AbstractI2CDevice
    {
        /* Measurement of pressure and temperature
         * 1)   Start temperature measurement
         * 2)   Wait 4.5ms
         * 3)   Read UT
         * 4)   Start pressure measurement
         * 5)   wait (depends on mode)
         * 6)   Read UP
         * 7)   Calculate pressure and temperarure in physical units
         */

        #region enums
        public enum OverSampling { UltraLowPower = 0, Standard = 1, HighResolution = 2, UltraHighResolution = 3 };
        public enum Register : byte
        {
            BMP085_PROM_START__ADDR = 0xAA,
            BMP085_PROM_DATA__LEN = 22,
            BMP085_CHIP_ID_REG = 0xD0,
            BMP085_VERSION_REG = 0xD1,
            BMP085_CTRL_MEAS_REG = 0xF4,
            BMP085_ADC_OUT_MSB_REG = 0xF6,
            BMP085_ADC_OUT_LSB_REG = 0xF7,
            BMP085_SOFT_RESET_REG = 0xE0,
            BMP085_T_MEASURE = 0x2E,				// temperature measurent 
            BMP085_P_MEASURE = 0x34             // pressure measurement	
        }
        #endregion

        #region Properties
        public double Temperature { get; private set; }
        public double Pressure { get; private set; }
        public double AbsoluteAltitude { get; private set; }
        public double SeaLevelPressure { get; set; }
        #endregion

        #region fields
        private OverSampling _overSampling = OverSampling.Standard;
        private Int16 AC1, AC2, AC3, B1, B2, MB, MC, MD;
        private UInt16 AC4, AC5, AC6;
        #endregion

        #region constructor
        /// <summary>
        /// 
        /// </summary>
        /// <param name="address">Usually 0x77</param>
        /// <param name="clockRate"></param>
        /// <param name="timeout"></param>
        public BMP085(ushort address, int clockRate, int timeout, OverSampling overSampling)
            : base(address, clockRate, timeout)
        {
            Temperature = 0;
            Pressure = 0;
            AbsoluteAltitude = 0;
            _overSampling = overSampling;

            if (Connected())
            {
                // Calibration
                I2CDevice.I2CTransaction[] xActions = new I2CDevice.I2CTransaction[2];
                xActions[0] = I2CDevice.CreateWriteTransaction(new byte[] { (byte)Register.BMP085_PROM_START__ADDR });
                byte[] RawData = new byte[22];
                xActions[1] = I2CDevice.CreateReadTransaction(RawData);
                int b = Write(xActions, 100);
                if (b != 23)
                    throw new ArgumentException("Error reading calibration coefficients");
                AC1 = (short)((short)(RawData[0] << 8) + (short)(RawData[1]));
                AC2 = (short)((short)(RawData[2] << 8) + (short)(RawData[3]));
                AC3 = (short)((short)(RawData[4] << 8) + (short)(RawData[5]));
                AC4 = (ushort)((ushort)(RawData[6] << 8) + (ushort)(RawData[7]));
                AC5 = (ushort)((ushort)(RawData[8] << 8) + (ushort)(RawData[9]));
                AC6 = (ushort)((ushort)(RawData[10] << 8) + (ushort)(RawData[11]));
                B1 = (short)((short)(RawData[12] << 8) + (short)(RawData[13]));
                B2 = (short)((short)(RawData[14] << 8) + (short)(RawData[15]));
                MB = (short)((short)(RawData[16] << 8) + (short)(RawData[17]));
                MC = (short)((short)(RawData[18] << 8) + (short)(RawData[19]));
                MD = (short)((short)(RawData[20] << 8) + (short)(RawData[21]));
                if (AC1 * AC2 * AC3 * AC4 * AC5 * AC6 * B1 * B2 * MB * MC * MD == 0)
                    throw new Exception("Error checking calibration coefficients");
            }
            else
            {
                throw new Exception("BMP085 is not connected");
            }
        }
        #endregion

        #region public
        public override bool Connected()
        {
            byte[] data = DeviceIdentifier();
            return (data[0] == 0x55);
        }

        public override byte[] DeviceIdentifier()
        {
            // returns =?
            return Read((byte)Register.BMP085_CHIP_ID_REG);
        }

        public bool CalculateValues()
        {
            double X1, X2, B5, B6, X3, B3, P, UT, UP;
            double B4, B7;

            //1)   Start temperature measurement
            I2CDevice.I2CTransaction[] xActions = new I2CDevice.I2CTransaction[1];
            xActions[0] = I2CDevice.CreateWriteTransaction(new byte[] { (byte)Register.BMP085_CTRL_MEAS_REG, (byte)Register.BMP085_T_MEASURE });
            if (Write(xActions, 100) != 2)
                return false;

            //2)    Wait 4.5ms
            Thread.Sleep(5);

            //3)   Read UT
            I2CDevice.I2CTransaction[] xActions2 = new I2CDevice.I2CTransaction[2];
            xActions2[0] = I2CDevice.CreateWriteTransaction(new byte[] { (byte)Register.BMP085_ADC_OUT_MSB_REG });
            byte[] RawData = new byte[2];
            xActions2[1] = I2CDevice.CreateReadTransaction(RawData);
            if (Write(xActions2, 100) != 3)
                return false;
            UT = (Int32)((Int32)(RawData[0] << 8) + (Int32)(RawData[1]));

            //4)   Start pressure measurement
            xActions[0] = I2CDevice.CreateWriteTransaction(new byte[] { (byte)Register.BMP085_CTRL_MEAS_REG, (byte)(((int)_overSampling << 6) + (byte)Register.BMP085_P_MEASURE) });
            if (Write(xActions, 100) != 2)
                return false;

            //5)   wait (depends on mode)
            Thread.Sleep(GetWaitTime());

            //6)   Read UP
            xActions2[0] = I2CDevice.CreateWriteTransaction(new byte[] { (byte)Register.BMP085_ADC_OUT_MSB_REG });
            byte[] RawData2 = new byte[3];
            xActions2[1] = I2CDevice.CreateReadTransaction(RawData2);
            if (Write(xActions2, 100) != 4)
                return false;
            UP = (((Int32)(RawData2[0]) << 16) + ((Int32)(RawData2[1]) << 8) + (Int32)RawData2[2]) >> (8 - (int)_overSampling);

            //7)   Calculate pressure and temperarure in physical units
            X1 = ((UT - AC6) * AC5) / System.Math.Pow(2, 15);
            X2 = (MC * System.Math.Pow(2, 11)) / (X1 + MD);
            B5 = X1 + X2;
            Temperature = (B5 + 8) / System.Math.Pow(2, 4);

            // Temperature is calculated in 0.1C units, so divide by 10
            Temperature = Temperature / 10;

            B6 = B5 - 4000;
            X1 = (AC2 * B6) / System.Math.Pow(2, 11);
            X2 = (B2 * ((B6 * B6) / System.Math.Pow(2, 12))) / System.Math.Pow(2, 11);
            X3 = X1 + X2;
            B3 = (Int32)(
                (
                (int)(AC1 * 4 + X3) << ((int)_overSampling + 2) / 4)
                );
            X1 = (AC3 * B6) / System.Math.Pow(2, 13);
            X2 = (B1 * (B6 * B6 / System.Math.Pow(2, 12)) / System.Math.Pow(2, 16));
            X3 = ((X1 + X2) + 2) / System.Math.Pow(2, 2);
            B4 = AC4 * (UInt32)(X3 + 32768) / System.Math.Pow(2, 15);
            B7 = ((UInt32)(UP - B3)) * (UInt32)(50000 >> (int)_overSampling);
            P = (B7 < 0x80000000) ? (Int32)((B7 * 2) / B4) : (Int32)((B7 / B4) * 2);
            X1 = (P / System.Math.Pow(2, 8)) * (P / System.Math.Pow(2, 8));
            X1 = (X1 * 3038) / System.Math.Pow(2, 16);
            X2 = (-7357 * P) / System.Math.Pow(2, 16);
            Pressure = P + ((X1 + X2 + 3791) / System.Math.Pow(2, 4));
            AbsoluteAltitude = (44330.75 * (1 - System.Math.Pow((double)Pressure / SeaLevelPressure, (1 / 5.255))));

            xActions = null;
            xActions2 = null;
            return true;
        }
        #endregion


        #region private
        /// <summary>
        /// Wait Time depinding on the OverSampling
        /// </summary>
        /// <returns></returns>
        private byte GetWaitTime()
        {
            switch (_overSampling)
            {
                case OverSampling.UltraLowPower:
                    return 5;
                case OverSampling.Standard:
                    return 8;
                case OverSampling.HighResolution:
                    return 14;
                case OverSampling.UltraHighResolution:
                    return 24;
            }
            return 24;
        }
        #endregion


    }
}
