/* ===================================================================================
 * Class to interface with HMC5883L sensor boards/chips
 * ===================================================================================
 * Author: Thomas D. Kryger
 * ===================================================================================
  * You may use, copy, borrow, modify or do anything you like with this code. Crediting
 * the author will be appreciated.
 * ===================================================================================
 */
using System;
using LoveElectronics.Resources;
using NETMF.IO.I2C;
using SensorTest.Tools;

namespace SensorTest.SensorDrivers
{
    /// <summary>
    /// This is Honeywell's HMC5883L, a 3-axis digital magnetometer designed for low-field magnetic sensing. The sensor has a full-scale range of 8 Guass and a resolution of up to 5 milli-Gauss.
    /// 
    /// I'm currently only using the HMC5883L Continous Measurement mode, so i have not (yet) implemented Single Measurement mode
    /// Datasheet: http://dlnmh9ip6v2uc.cloudfront.net/datasheets/Sensors/Magneto/HMC5883L-FDS.pdf
    /// </summary>
    public class HMC5883L : AbstractI2CDevice
    {
        #region Registers and Modes
        public enum Register : byte
        {
            ScaleRegister = 0x01,
            MeasurementRateRegister = 0x02,
            DataValue_StartingRegister = 0x03,
            IdentificationRegister_1 = 0x0A,
            IdentificationRegister_2 = 0x0B,
            IdentificationRegister_3 = 0x0C,
            SelfTest = 0xBB
        }

        public enum IdentificationValues : byte
        {
            A_VALUE = 0x48,
            B_VALUE = 0x34,
            C_VALUE = 0x33
        }

        #region Configuation Register A
        public enum DataOutputRates : byte
        {
            Rate_075 = 0x00,
            Rate_150 = 0x01,
            Rate_300 = 0x02,
            Rate_750 = 0x03,
            Rate_1500 = 0x04,
            Rate_3000 = 0x05,
            Rate_7500 = 0x06,
            Rate_reserved = 0x07
        }

        public enum MeasurementMode : byte
        {
            Normal = 0x00,
            PositiveBias = 0x01,
            NegativeBias = 0x02,
            reserved = 0x03
        }
        #endregion
        #region Configuration Register B
        public enum Gain : byte
        {
            Ga_088 = 0x00,
            Ga_130 = 0x01,
            Ga_190 = 0x02,
            Ga_250 = 0x03,
            Ga_400 = 0x04,
            Ga_470 = 0x05,
            Ga_560 = 0x06,
            Ga_810 = 0x07
        }
        #endregion

        /// <summary>
        /// Values are divided by 100 on use..
        /// </summary>
        public enum GaussScale : int
        {
            Gauss0p88 = 88,
            Gauss1p3 = 130,
            Gauss1p9 = 190,
            Gauss2p5 = 250,
            Gauss4 = 400,
            Gauss4p7 = 470,
            Gauss5p6 = 560,
            Gauss8p1 = 810
        }

        public enum OperatingMode : byte
        {
            /// <summary>
            /// Continuous-Measurement Mode. In continuous-measurement mode, the device continuously performs measurements and places the result in the data register. RDY goes high when new data is placed in all three registers. After a power-on or a write to the mode or configuration register, the first measurement set is available from all three data output registers after a period of 2/fDO and subsequent measurements are available at a frequency of fDO, where fDO is the frequency of data output.
            /// </summary>
            ContinousMeasurement = 0x00,
            /// <summary>
            /// Single-Measurement Mode (Default). When single-measurement mode is selected, device performs a single measurement, sets RDY high and returned to idle mode. Mode register returns to idle mode bit values. The measurement remains in the data output register and RDY remains high until the data output register is read or another measurement is performed.
            /// </summary>
            SingleMeasurement = 0x01,
            /// <summary>
            /// Idle Mode. Device is placed in idle mode.
            /// </summary>
            IdleMode10 = 0x10,
            /// <summary>
            /// Idle Mode. Device is placed in idle mode.
            /// </summary>
            IdleMode11 = 0x11
        }
        #endregion

        #region Three Axis Operations
        public ThreeAxisRawData Raw
        {
            get
            {
                byte[] bytes = this.Read((byte)Register.DataValue_StartingRegister, 6);

                short xReading = (short)((bytes[0] << 8) | bytes[1]);
                short zReading = (short)((bytes[2] << 8) | bytes[3]);
                short yReading = (short)((bytes[4] << 8) | bytes[5]);

                _rawData.X = xReading;
                _rawData.Y = yReading;
                _rawData.Z = zReading;

                return _rawData;
            }
        }

        public ThreeAxisScaledData ScaledData
        {
            get
            {
                _scaledData.ScaledX = Scale * (double)this.Raw.X;
                _scaledData.ScaledY = Scale * (double)this.Raw.Y;
                _scaledData.ScaledZ = Scale * (double)this.Raw.Z;
                return _scaledData;
            }
        }
        #endregion

        #region Fields
        private ThreeAxisRawData _rawData = new ThreeAxisRawData();
        private ThreeAxisScaledData _scaledData = new ThreeAxisScaledData();
        #endregion

        #region Constructor
        public HMC5883L(ushort address, int clockRate, int timeout)
            : base(address, clockRate, timeout)
        {
            SetScale(GaussScale.Gauss1p3);
            SetMeasurementMode(OperatingMode.ContinousMeasurement);
            Pitch = 0;
            Roll = 0;
        }
        #endregion

        #region Properties
        /// <summary>
        /// Pitch (Y) i radians. Set to value from accelerometer to get tilt compensated heading
        /// </summary>
        public double Pitch { get; set; }
        /// <summary>
        /// Roll (X) i radians. Set to value from accelerometer to get tilt compensated heading
        /// </summary>
        public double Roll { get; set; }
        public double Scale { get; private set; }
        public double HeadingRaw
        {
            get
            {
                double y = ScaledData.ScaledY;
                double x = ScaledData.ScaledX;
                if (x != 0)
                    return ExMath.Atan(y / x);
                else
                    return 0;
            }
        }
        public double HeadingRawDegrees() { return ExMath.RadiansToDegrees(this.HeadingRaw); }
        public double HeadingDegrees() { return ExMath.RadiansToDegrees(this.GetHeading()); }
        #endregion

        #region public methods
        public void SetScale(GaussScale gaussValue)
        {
            byte regValue = 0x00;
            double gauss = (double)gaussValue / 100;
            if (gauss == 0.88)
            {
                regValue = 0x00;
                Scale = 0.73;
            }
            else if (gauss == 1.3)
            {
                regValue = 0x01;
                Scale = 0.92;
            }
            else if (gauss == 1.9)
            {
                regValue = 0x02;
                Scale = 1.22;
            }
            else if (gauss == 2.5)
            {
                regValue = 0x03;
                Scale = 1.52;
            }
            else if (gauss == 4.0)
            {
                regValue = 0x04;
                Scale = 2.27;
            }
            else if (gauss == 4.7)
            {
                regValue = 0x05;
                Scale = 2.56;
            }
            else if (gauss == 5.6)
            {
                regValue = 0x06;
                Scale = 3.03;
            }
            else if (gauss == 8.1)
            {
                regValue = 0x07;
                Scale = 4.35;
            }
            else
            {
                throw new ArgumentException("Unknown gauss value");
            }

            regValue = (byte)(((int)regValue) << 5);

            this.Write((byte)Register.ScaleRegister, regValue);
        }


        // Since SingleMeasurement is not implemented, this could cause problems, so its private for now
        private void SetMeasurementMode(OperatingMode mode)
        {
            this.Write((byte)Register.MeasurementRateRegister, (byte)mode);
        }

        /// <summary>
        /// TODO GetReadings()
        /// Get new readings when running in standard mode.
        /// Only call it every 60 ms
        /// </summary>
        //public void GetReadings()
        //{
            
        //}

        public double GetHeading()
        {
            ThreeAxisScaledData sd;
            double Xm, Ym, Zm;
            double Xh, Yh;
            double cp, sp, cr, sr; // Cos and Sin of Pitch and Roll

            sd = this.ScaledData; // this ensures we are only reading once from the bus
            Xm = sd.ScaledX;
            Ym = sd.ScaledY;
            Zm = sd.ScaledZ;

            cp = ExMath.Cos(Pitch);
            sp = ExMath.Sin(Pitch);
            cr = ExMath.Cos(Roll);
            sr = ExMath.Sin(Roll);

            Xh = Xm * cp + Zm * sp;
            Yh = Xm * sr * sp + Ym * cr - Zm * sr * cp;

            if (Xh != 0)
                return ExMath.Atan(Yh / Xh);
            else
                return 0;
        }
        #endregion
        public override bool Connected()
        {
            return ((DeviceIdentifier()[0] == (byte)IdentificationValues.A_VALUE) &&
                (DeviceIdentifier()[1] == (byte)IdentificationValues.B_VALUE) &&
                (DeviceIdentifier()[2] == (byte)IdentificationValues.C_VALUE));
        }

        public override byte[] DeviceIdentifier()
        {
            byte[] d =new byte[3] { Read((byte)Register.IdentificationRegister_1)[0],
                Read((byte)Register.IdentificationRegister_2)[0],
                Read((byte)Register.IdentificationRegister_3)[0] };
            return d;
        }
    }
}
