/* ===================================================================================
 * Class to interface with ITG-3200 sensor boards/chips
 * ===================================================================================
 * Author: Thomas D. Kryger
 * ===================================================================================
  * You may use, copy, borrow, modify or do anything you like with this code. Crediting
 * the author will be appreciated.
 * ===================================================================================
 */
using System;
using System.Threading;
using Microsoft.SPOT.Hardware;
using NETMF.IO.I2C;
using SensorTest.Tools;

namespace SensorTest.SensorDrivers
{
    /// <summary>
    /// 3-Axis I2C interfaced Gyro library for ITG-3200 sensor
    /// Datasheet: http://invensense.com/mems/gyro/documents/PS-ITG-3200A.pdf
    /// </summary>
    public class ITG3200 : AbstractI2CDevice
    {
        #region Delegates and events
        public delegate void ITG3200NewData(uint data1, uint data2, DateTime time);
        public event ITG3200NewData OnNewData;

        void interruptPort_OnInterrupt(uint data1, uint data2, DateTime time)
        {
            if (OnNewData != null)
                OnNewData(data1, data2, time);
        }
        #endregion

        #region Consts
        //Scaling factor - 1LSB in raw output = 1/14 degrees/sec
        const double scaleFactor = 14.375;
        #endregion

        #region fields
        private InterruptPort interruptPort;
        private ThreeAxisRawData _rawData = new ThreeAxisRawData();
        private double xOffset = 0;
        private double yOffset = 0;
        private double zOffset = 0;
        #endregion

        #region ITG3200 Registers
        /// <summary>
        /// An enum providing descriptive names for the gyro registers
        /// </summary>
        public enum Register : byte
        {
            WHO_AM_I = 0x00,
            SMPLRT_DIV = 0x15,
            DLPF_FS = 0x16,
            INTERRUPT = 0x17,
            GYRO_TOUT_H = 0x1B,
            GYRO_TOUT_L = 0x1C,
            GYRO_XOUT_H = 0x1D,
            GYRO_XOUT_L = 0x1E,
            GYRO_YOUT_H = 0x1F,
            GYRO_YOUT_L = 0x20,
            GYRO_ZOUT_H = 0x21,
            GYRO_ZOUT_L = 0x22,
            PWR_MGMT = 0x3E,
            INT_STATUS = 0x1A,
        }

        /// <summary>
        /// An enum providing descriptive names for the combined Lowpass filter and Internal sample rate register
        /// </summary>
        public enum Filter : byte
        {
            LP256Hz_Int8kHz = 0x30,
            LP188Hz_Int1kHz = 0x32,
            LP98Hz_Int1kHz = 0x36,
            LP42Hz_Int1kHz = 0x38,
            LP20Hz_Int1kHz = 0x3A,
            LP10Hz_Int1kHz = 0x3C,
            LP5Hz_Int1kHz = 0x3E,
        }

        /// <summary>
        ///  An enum providing descriptive names for the various clock sources
        /// </summary>
        public enum ClockSource : byte
        {
            INT_OSC = 0x0,
            PLL_XGRYO_REF = 0x1,
            PLL_YGRYO_REF = 0x2,
            PLL_ZGRYO_REF = 0x3,
            PLL_EXT_32kHz = 0x4,
            PLL_EXT_19MHz = 0x5,
        }
        #endregion

        #region Constructor
        public ITG3200(ushort address, int clockRate, int timeout, Cpu.Pin InterruptPort)
            : base(address, clockRate, timeout)
        {
            if (!Connected())
                throw new Exception("ITG3200 not connected");
            if (InterruptPort != Cpu.Pin.GPIO_NONE)
            {
                interruptPort = new InterruptPort(InterruptPort, true, Port.ResistorMode.Disabled, Port.InterruptMode.InterruptEdgeHigh);
            }
            SampleRateDivider = 40;
            LowpassFilterCutoff = Filter.LP20Hz_Int1kHz;
            Sleep(false, false, false, false);
            SetClockSource(ClockSource.PLL_ZGRYO_REF);
            Thread.Sleep(100);
            Calibrate();
        }
        #endregion

        #region Properties
        public ThreeAxisRawData RawData 
        { 
            get 
            {
                byte[] bytes = this.Read((byte)Register.GYRO_XOUT_H, 6);

                short xReading = (short)((bytes[0] << 8) | bytes[1]);
                short zReading = (short)((bytes[2] << 8) | bytes[3]);
                short yReading = (short)((bytes[4] << 8) | bytes[5]);

                _rawData.X = xReading;
                _rawData.Y = yReading;
                _rawData.Z = zReading;
                return _rawData; 
            } 
        }
        public byte SampleRateDivider
        {
            get
            {
                return Read((byte)Register.SMPLRT_DIV)[0];
            }
            set
            {
                Write((byte)Register.SMPLRT_DIV, value);
            }
        }

        public Filter LowpassFilterCutoff
        {
            get
            {
                return (Filter)Read((byte)Register.DLPF_FS)[0];
            }
            set
            {
                Write((byte)Register.DLPF_FS, (byte)value);
            }
        }

        #region Axis data and Temp reading
        /// <summary>
        /// Read the X Gyro (horizontal, parallel to text on device)
        /// </summary>
        /// <returns>degrees/sec</returns>
        public double X
        {
            get
            {
                int data = ReadShort((byte)Register.GYRO_XOUT_H, (byte)Register.GYRO_XOUT_L);
                return (((double)(data)) / scaleFactor) - xOffset;
            }
        }


        /// <summary>
        /// Read the Y Gyro (horizontal, perpendicular to text on device)
        /// </summary>
        /// <returns>degrees/sec</returns>
        public double Y
        {
            get
            {
                int data = ReadShort((byte)Register.GYRO_YOUT_H, (byte)Register.GYRO_YOUT_L);
                return (((double)(data)) / scaleFactor) - yOffset;
            }
        }


        /// <summary>
        /// Read the Z Gyro (vertical, looking down)
        /// </summary>
        /// <returns>degrees/sec</returns>     
        public double Z
        {
            get
            {
                int data = ReadShort((byte)Register.GYRO_ZOUT_H, (byte)Register.GYRO_ZOUT_L);
                return (((double)(data)) / scaleFactor) - zOffset;
            }
        }


        /// <summary>
        /// Read the temperature sensor
        /// </summary>
        /// <returns>Current temperature, in degrees C</returns>
        public int Temperature
        {
            get
            {
                int data = ReadShort((byte)Register.GYRO_TOUT_H, (byte)Register.GYRO_TOUT_L);
                //Offset = -35 degrees, 13200 counts. 280 counts/degrees C.
                return 35 + ((data + 13200) / 280);
            }
        }
        #endregion
        #endregion

        #region public methods
        /// <summary>
        /// Calibrate the offsets. Ensure the Gyro is not moving during this function
        /// </summary>
        public void Calibrate()
        {
            int i = 0;
            double newxOffset = 0;
            double newyOffset = 0;
            double newzOffset = 0;

            for (i = 0; i < 20; i++)
            {
                Thread.Sleep(10);
                newxOffset += X;
                newyOffset += Y;
                newzOffset += Z;
            }

            newxOffset /= i;
            newyOffset /= i;
            newzOffset /= i;

            xOffset = newxOffset;
            yOffset = newyOffset;
            zOffset = newzOffset;

            //enable interrupts now that we're calibrated
            if (interruptPort != null)
            {
                interruptPort.OnInterrupt += new NativeEventHandler(interruptPort_OnInterrupt);
            }
        }



        /// <summary>
        /// Set the clock source of the gyro
        /// </summary>
        /// <param name="ClkSource">The source of the clock</param>
        public void SetClockSource(ClockSource ClkSource)
        {
            byte curStatus = Read((byte)Register.PWR_MGMT)[0];
            curStatus |= (byte)ClkSource;

            Write((byte)Register.PWR_MGMT, curStatus);
        }

        /// <summary>
        /// Set the sleep mode of the gyro, reducing power usage
        /// </summary>
        /// <param name="isSleep">Power down the entire device, except serial interface and registers</param>
        /// <param name="xSleep">Set the X gyro to standby</param>
        /// <param name="ySleep">Set the Y gyro to standby</param>
        /// <param name="zSleep">Set the Z gyro to standby</param>
        public void Sleep(bool isSleep, bool xSleep, bool ySleep, bool zSleep)
        {
            byte curStatus = Read((byte)Register.PWR_MGMT)[0];

            if (isSleep)
                curStatus |= 1 << 6;
            else
                curStatus &= 0 << 6;

            if (xSleep)
                curStatus |= 1 << 5;
            else
                curStatus &= 0 << 5;

            if (ySleep)
                curStatus |= 1 << 4;
            else
                curStatus &= 0 << 4;

            if (zSleep)
                curStatus |= 1 << 3;
            else
                curStatus &= 0 << 3;

            Write((byte)Register.PWR_MGMT, curStatus);
        }

        /// <summary>
        /// Reset the gyro's registers to their default values. This includes disabling interrupts
        /// </summary>
        public void Reset()
        {
            byte curStatus = Read((byte)Register.PWR_MGMT)[0];
            curStatus |= 1 << 7;
            Write((byte)Register.PWR_MGMT, curStatus);
        }
        #region Settings
        /// <summary>
        /// Enable interrupts (INT pin) on the gyro
        /// </summary>
        public void EnableInterrupts()
        {
            Write((byte)Register.INTERRUPT, 0x11);
        }

        /// <summary>
        /// Disable interrupts (INT pin) on the gyro
        /// </summary>
        public void DisableInterrupts()
        {
            Write((byte)Register.INTERRUPT, 0x0);
        }
        #endregion
        #endregion

        public override bool Connected()
        {
            byte me = DeviceIdentifier()[0];
            // ITG3200 has 2 device id 
            return ((me == 0x68) || (me == 0x69));
        }

        public override byte[] DeviceIdentifier()
        {
            return new byte[1] { Read((byte)Register.WHO_AM_I)[0] };
        }
    }
}
